#include <Windows.h>
#include <WinLogic.h>
#include <stdio.h>
#include "MMC.h"

short port = 888;

void MMC_cs(char cs) {
	Out32(port, Inp32(port) & ~1 | !!cs);
}

void MMC_data(char d) {
	Out32(port, Inp32(port) & ~4 | (!!d)<<2);
}

char MMC_get() {
	return !pp_busy(port);
}

void MMC_clock(char c) {
	Out32(port, Inp32(port) & ~2 | (!!c)<<1);
}

BYTE bitswap(BYTE byte){
	BYTE temp=0;
	for(int i=0;i<8;i++){
		temp=temp<<1;
		temp=temp | (byte & 1);
		byte=byte>>1;
	}
	return temp;
}

//SendByte - in SPI MASTER fashion
//will READ a byte everytime it writes a byte
BYTE SendByte(BYTE byte) {
	//CLOCK is on pin 4 of LPT1-output
	//MISO is on pin 11 of LPT1 -input
	//MOSI is on pin 3 of LPT1  -output
	//SS is pin 5 of LPT1 - output
	char bit;
	BYTE input=0,bitin;
	int i;
	byte=bitswap(byte);

	//get first bit to send
	for(i=0;i<8;i++){
		bit=byte & 1;
		byte = byte >> 1;
		//lower CLOCK
		MMC_clock(0);
		//put DATA on MOSI line
		MMC_data(bit);
		//raise CLOCK
		MMC_clock(1);
		bitin=MMC_get();
		input|=bitin;
		if(i!=7) input<<=1;    
	}
	//Raise Data - acting like a pull-up resistor
	MMC_data(1);
	return input;
}

void SendCmd(BYTE command, UINT argument, BYTE CRC){
  BYTE cmd[6];
  BYTE temp;
  int i;
  cmd[0]=(command|0x40);
  for(i=3;i>=0;i--){
    temp=(BYTE)(argument>>(8*i));
    cmd[4-i]=(temp);
  }
  cmd[5]=(CRC);
  for(i=0;i<6;i++)
    SendByte(cmd[i]);
}

BYTE MMCgetResponse(){
  //Response comes 1-8bytes after command
  //the first bit will be a 0
  //followed by an error code
  //data will be 0xff until response
  int i=0;
  BYTE response;
  
  while(i<=8){
    response=SendByte(0xff);
    if(response==0x00)
      break;
    if(response==0x01)
      break;
    i++;
  }
  return response;
}

bool InitMMC(void){
	//raise SS and MOSI for 80 clock cycles
	//SendByte(0xff) 10 times with SS high
	//RAISE SS
	int delay=0, trials=0;
	char response=0x01;
	//initialization sequence on PowerUp
	MMC_cs(1);
	for(int i=0;i<=9;i++)
		SendByte(0xff);
	MMC_cs(0);  
	//Send Command 0 to put MMC in SPI mode
	SendCmd(0x00,0,0x95);
	//Now wait for READY RESPONSE
	//response=MMCgetResponse();
	if(MMCgetResponse()!=0x01)
		return false;

	while(response==0x01){
		printf("Sending Command 1\n");
		MMC_cs(1);
		SendByte(0xff);
		MMC_cs(0);
		SendCmd(0x01,0x00ffc000,0xff);
		response=MMCgetResponse();
	} 
	if(response==0x00)
		printf("RESPONSE WAS GOOD\n");

	MMC_cs(1);
	SendByte(0xff);

	printf("MMC INITIALIZED AND SET TO SPI MODE PROPERLY.\n");
	return true;  
}

bool MMCwriteBlock(char *Block, int address){
  char busy,dataResp;
  //Block size is 512 bytes exactly
  //First Lower SS
  MMC_cs(0);
  //Then send write command
  SendCmd(24,address,0xff);
  if(MMCgetResponse()==00){
    //command was a success - now send data
    //start with DATA TOKEN = 0xFE
    SendByte(0xfe);
    //now send data
    for(int count=0;count<512;count++){
      SendByte(Block[count]);
    }
    //data block sent - now send checksum
    SendByte(0xff);
    SendByte(0xff);
    //Now read in the DATA RESPONSE token
    dataResp=SendByte(0xff);
    //Following the DATA RESPONSE token
    //are a number of BUSY bytes
    //a zero byte indicates the MMC is busy
    busy=SendByte(0xff);
    while(busy==0){
      busy=SendByte(0xff);
    }
    dataResp=dataResp&0x0f;	//mask the high byte of the DATA RESPONSE token
    MMC_cs(1);
    SendByte(0xff);
    if(dataResp==0x0b){
      printf("DATA WAS NOT ACCEPTED BY CARD -- CRC ERROR\n");
      return false;
      }
    if(dataResp==0x05)
      return true;
      
    printf("Invalid data Response token.\n");
    return false;
 }
  printf("Command 24 (Write) was not received by the MMC.\n");
  return false;
}

bool MMCreadBlock(void *mem, UINT address, UINT size) {
	UINT a;

	MMC_cs(0);
	SendCmd(16, size, 0xFF);
	if (MMCgetResponse()) return false;
	SendByte(0xFF);

	SendCmd(17, address, 0xFF);
	if (MMCgetResponse()) return false;

	while (SendByte(0xFF) != 0xFE);
	for (a=0;a<size;++a)
		*((BYTE*)mem+a) = SendByte(0xFF);
	SendByte(0xFF);	// dozna eksumu...
	SendByte(0xFF);

	MMC_cs(1);
	SendByte(0xFF);

	return true;
}

#define SZ 512

int main() {
	int a;
	char cmd[100];
	printf("Hello!\n");
	do {
		gets(cmd);
		if (!strncmp(cmd, "cs", 2)) {
			MMC_cs(atoi(cmd+2));
		} else if (!strncmp(cmd, "data", 4)) {
			MMC_data(atoi(cmd+4));
		} else if (!strncmp(cmd, "get", 3)) {
			printf("Dobio: %d\n", MMC_get());
		} else if (!strncmp(cmd, "clock", 5)) {
			MMC_clock(atoi(cmd+5));
		} else if (!strcmp(cmd, "init")) {
			printf("Uspjeh: %d\n", InitMMC());
		} else if (!strcmp(cmd, "read")) {
			BYTE data[SZ];
			MMCreadBlock(data, 0, SZ);
			for (a=0;a<SZ;++a) {
				printf("%02X", data[a]);
			}
			printf("\n");
		} else if (!strncmp(cmd, "file ", 5)) {
			int n;
			FILE *f;
			BYTE d[512];

			printf("Koliko blokova? ");
			scanf("%d", &n);
			f = fopen(cmd+5, "wb");
			for (a=0;a<n;++a) {
				printf("Reading block %d/%d...\n", a+1,n);
				if (!MMCreadBlock(d, a*512, 512)) {
					printf("Greka!\n");
					break;
				}
				fwrite(d, 512, 1, f);
			}
			fclose(f);
			printf("Done!\n");
		} else if (!strcmp(cmd, "resp")) {
			printf("response: 0x%02X\n", MMCgetResponse());
		}
	} while (stricmp(cmd, "exit"));
	return 0;
}
